#include "common.h"
#include "LtrCrateThreads.h"

/*------------------------------------------------------------------------------------------------*/
/*    exception */
#define FATAL true
#define NONFATAL false
class Ex
    {
private:
    const char *errmsg;
    AnsiString task;
    int level;
    int errcode;
    bool is_fatal;
public:
    Ex(bool fatal, const char *err_msg, const AnsiString &log_task, int log_level, int err_code = 0)
        {
        is_fatal = fatal;
        errmsg = err_msg; task = log_task; level = log_level; errcode = err_code;
        }

    Ex()
        {
        errmsg = NULL; errcode = 0; task = ""; level = 0; is_fatal = false;
        }

    void Dispatch(const char *host_id)
        {
        if (!errmsg) return;
        if (errcode)
            {
            LOGWRITE(Format("%s: %s (error %d)",
                ARRAYOFCONST((host_id, errmsg, errcode))), task, level);
            }
        else
            {
            LOGWRITE(Format("%s: %s",
                ARRAYOFCONST((host_id, errmsg))), task, level);
            }
        }

    bool IsFatal(void)
        {
        return is_fatal;
        }
    };
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
__fastcall TReadUSBCrateThread::TReadUSBCrateThread(TCrateUSB *_crate, bool CreateSuspended)
    : TThread(true)
    {
    crate = _crate;
    Priority = tpHigher;
    FreeOnTerminate = false;
    init_complete = CreateEvent(NULL, EVT_MANUAL_RESET, EVT_NONSIGNALED, NULL);
    if (!CreateSuspended) Resume();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
__fastcall TReadUSBCrateThread::~TReadUSBCrateThread(void)
    {
    CloseHandle(init_complete);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void __fastcall TReadUSBCrateThread::Execute(void)
    {
    bool failed = false;

    DWORD CrateReadSize = CRATE_RR_SIZE_MAX_DWORD * sizeof(DWORD);

    switch (crate->usb_info.speed)
        {
        case USB_Speed_Low:
        case USB_Speed_Full:
            CrateReadSize /= 16;
            break;
        }

    memset(rreq, 0, sizeof(rreq));
    for (int i = 0; i < CRATE_RR_N; i++)
        {
        //          ,   DWORD
        rreq[i].buf = &dwReadBuf[i][1];
        if ((rreq[i].over.hEvent = CreateEvent(NULL, EVT_MANUAL_RESET, EVT_SIGNALED, NULL)) == NULL)
            {
            LOGWRITE(Format("%s: Read thread error allocating OVERLAPPED event",
                ARRAYOFCONST((crate->name))),
                LOGTASK_CRATE_READ, LOGL_ERROR);
            failed = true;
            break;
            }
        }

    //       
    for (int i = 0; i < CRATE_RR_N; i++)
        {
        DWORD dummy;
        if (!ReadFile(crate->DeviceHANDLE, (BYTE*)rreq[i].buf, CrateReadSize, &dummy, &rreq[i].over))
            {
            int err = GetLastError();
            if (err != ERROR_IO_PENDING)
                {
                failed = true;
                LOGWRITE(Format("%s: USB read request failed (error %d)",
                    ARRAYOFCONST((crate->name, err))),
                    LOGTASK_CRATE_READ, LOGL_ERROR);
                break;
                }
            }
        }

#if DEBUG_PROFILER
    Prf::Add _(3, true);
#endif        
    SetEvent(init_complete); //   
    DWORD index = 0;
    HANDLE evt[2];
    DWORD XferSize = 0;
    DWORD RxOffset = 0; //  0..3     DWORD
    BYTE FixupBuf[sizeof(DWORD)]; //     DWORD
    crate->special_code = 0;
    crate->total_count_recv = 0;
    while (!failed)
        {
        evt[0] = rreq[index].over.hEvent;
        evt[1] = crate->hStopEvent;
        _P;
        DWORD wait_res = WaitForMultipleObjects(2, evt, FALSE, INFINITE);
        _R;
        if (wait_res != WAIT_OBJECT_0)
            { //   stop --  
            break;
            }

        if (!GetOverlappedResult(crate->DeviceHANDLE, &rreq[index].over, &XferSize, FALSE))
            {
            int err = GetLastError();
            LOGWRITE(Format("%s: USB connection lost (%d)",
                ARRAYOFCONST((crate->name, err))),
                LOGTASK_CRATE_READ, LOGL_INFO);
            failed = true;
            break;
            }

        /* XferSize =  ,   .
         * RxOffset =   ,   DWORD (  dwFixup).
         *   ,        
         * (   DWORD -- rreq[i].buf = &dwReadBuf[i][1]) */
        BYTE* data_ptr = (BYTE*)rreq[index].buf;
        for (int k = RxOffset - 1; k >= 0; k--)
            {
            *(--data_ptr) = FixupBuf[k];
            }
        XferSize += RxOffset; //   ,   data_ptr
        RxOffset = XferSize % sizeof(DWORD); //   

        if (XferSize >= sizeof(DWORD))
            {
            if (!crate->ParseDataFromCrate((DWORD*)data_ptr, XferSize / sizeof(DWORD)))
                {
                failed = true;
                break;
                }
            }

        /*  RxOffset       FixupBuf */
        for (int k = 0; k < RxOffset; k++)
            {
            FixupBuf[k] = data_ptr[XferSize - RxOffset + k];
            }

        //      -  
        if (!ReadFile(crate->DeviceHANDLE, (BYTE*)rreq[index].buf, CrateReadSize, &XferSize, &rreq[index].over))
            {
            XferSize = 0;
            int err = GetLastError();
            if (err != ERROR_IO_PENDING)
                {
                failed = true;
                LOGWRITE(Format("%s: USB read request failed (error %d)",
                    ARRAYOFCONST((crate->name, err))),
                    LOGTASK_CRATE_READ, LOGL_ERROR);
                }
            }
        index = (index + 1) % CRATE_RR_N;
        } /* while !failed */

    LOGWRITE(Format("%s: Read thread shutting down", ARRAYOFCONST((crate->name))),
        LOGTASK_CRATE_READ, LOGL_DEBUG2);

    SetEvent(crate->hStopEvent);

    crate->ltr_USB->CancelIO();
    for (int i = 0; i < CRATE_RR_N; i++)
        {
        DWORD dummy = 0;
        if (rreq[i].over.hEvent)
            {
            GetOverlappedResult(crate->DeviceHANDLE, &rreq[i].over, &dummy, TRUE);
            CloseHandle(rreq[i].over.hEvent);
            }
        }

    /*         */
    if (failed)
        crate->SetState(CRATE_ST_DEAD);
    LOGWRITE(Format("%s: Read thread shut down.", ARRAYOFCONST((crate->name))),
        LOGTASK_CRATE_READ, LOGL_DEBUG1);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
__fastcall TWriteUSBCrateThread::TWriteUSBCrateThread(TCrateUSB *_crate, bool CreateSuspended)
    : TThread(true)
    {
    crate = _crate;
    Priority = tpHigher; // tpNormal;
    FreeOnTerminate = false;
    init_complete = CreateEvent(NULL, EVT_MANUAL_RESET, EVT_NONSIGNALED, NULL);
    if (!CreateSuspended) Resume();
    }
//---------------------------------------------------------------------------
__fastcall TWriteUSBCrateThread::~TWriteUSBCrateThread(void)
    {
    CloseHandle(init_complete);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void __fastcall TWriteUSBCrateThread::Execute(void)
    {
    bool failed = false;

    memset(&wreq, 0, sizeof(wreq));
    wreq.buf = dwWriteBuf;
    if ((wreq.over.hEvent = CreateEvent(NULL, EVT_MANUAL_RESET, EVT_SIGNALED, NULL)) == NULL)
        {
        LOGWRITE(Format("%s: Write thread error allocating OVERLAPPED event",
            ARRAYOFCONST((crate->name))),
            LOGTASK_CRATE_WRITE, LOGL_ERROR);
        failed = true;
        }

#if DEBUG_PROFILER
    Prf::Add _(2, true);
#endif        
    SetEvent(init_complete); //   
    crate->total_count_send = 0;
    HANDLE evt[2];
    DWORD SendSize = 0;
    DWORD ReadyDWords;
    DWORD skip = 0;
    while (!failed)
        {
        //      
        evt[0] = crate->hClientPutEvent;
        evt[1] = crate->hStopEvent;
        for (;;)
            {
            ReadyDWords = crate->PrepareDataForCrate(wreq.buf, CRATE_WR_SIZE_MAX_DWORD, skip);
            if (ReadyDWords) break;
            //     
            _P;
            int wait_res = WaitForMultipleObjects(2, evt, FALSE, TIMEOUT_WAIT_CLIENTPUT);
            //@@
            //if (wait_res == WAIT_OBJECT_0) Sleep(5);
            _R;
            if ((wait_res != WAIT_OBJECT_0) && (wait_res != WAIT_TIMEOUT))
                break;
            }
        if (!ReadyDWords) break; //   stop event

        //  
        if (!WriteFile(crate->DeviceHANDLE, (BYTE*)wreq.buf, ReadyDWords * sizeof(DWORD), &SendSize, &wreq.over))
            {
            int err = GetLastError();
            if (err != ERROR_IO_PENDING)
                {
                failed = true;
                LOGWRITE(Format("%s: USB write request failed (error %d)",
                    ARRAYOFCONST((crate->name, err))),
                    LOGTASK_CRATE_WRITE, LOGL_ERROR);
                break;
                }
            }
        //  
        crate->total_count_send += ReadyDWords;

        evt[0] = wreq.over.hEvent;
        evt[1] = crate->hStopEvent;
        _P;
        DWORD wait_res = WaitForMultipleObjects(2, evt, FALSE, INFINITE);
        _R;
        if (wait_res != WAIT_OBJECT_0)
            { //   stop --  
            break;
            }

        if (!GetOverlappedResult(crate->DeviceHANDLE, &wreq.over, &SendSize, FALSE))
            {
            int err = GetLastError();
            LOGWRITE(Format("%s: USB connection lost (%d)",
                ARRAYOFCONST((crate->name, err))),
                LOGTASK_CRATE_WRITE, LOGL_INFO);
            failed = true;
            break;
            }
        } /* while !failed */

    LOGWRITE(Format("%s: Write thread shutting down", ARRAYOFCONST((crate->name))),
        LOGTASK_CRATE_WRITE, LOGL_DEBUG2);

    SetEvent(crate->hStopEvent);

    crate->ltr_USB->CancelIO();
    if (wreq.over.hEvent)
        {
        GetOverlappedResult(crate->DeviceHANDLE, &wreq.over, &SendSize, TRUE);
        CloseHandle(wreq.over.hEvent);
        }

    /*         */
    if (failed)
        crate->SetState(CRATE_ST_DEAD);
    LOGWRITE(Format("%s: Write thread shut down.", ARRAYOFCONST((crate->name))),
        LOGTASK_CRATE_WRITE, LOGL_DEBUG1);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
__fastcall TIPCratePingThread::TIPCratePingThread(TCrateIP *_crate, bool CreateSuspended,
    DWORD _timeout, int _attempts)
    : TThread(true)
    {
    Priority = tpIdle;
    FreeOnTerminate = false;
    crate = _crate;
    ping_timeout = _timeout;
    ping_attempts = _attempts;
    hTermEvent = CreateEvent(NULL, EVT_MANUAL_RESET, EVT_NONSIGNALED, NULL);
    if (!CreateSuspended) Resume();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
__fastcall TIPCratePingThread::~TIPCratePingThread(void)
    {
    CloseHandle(hTermEvent);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void __fastcall TIPCratePingThread::TerminateByEvent(void)
    {
    SetEvent(hTermEvent);
    TThread::Terminate();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void __fastcall TIPCratePingThread::Execute(void)
    {
    /*      (      ).
     *   IdleTimer.      NOP.
     *         ,   
     *    .
     */
    HANDLE wait_list[2];
    int failcount;

    wait_list[0] = hTermEvent;
    wait_list[1] = crate->GetIdleTimerHandle();
    if (!wait_list[1])
        return;
    while (!Terminated)
        {
        switch (WaitForMultipleObjects(2, wait_list, FALSE, INFINITE))
            {
            case WAIT_OBJECT_0: /* TERMINATE EVENT */
                return;
            case WAIT_OBJECT_0 + 1: /* IDLE TIMER */
                failcount = crate->PingAndGetFailCount(ping_timeout);
                if (failcount == 0)
                    {
                    #if 0
                    LOGWRITE(Format("Ping: %s is alive",
                        ARRAYOFCONST((crate->path))),
                        LOGTASK_CRATE_CTL, LOGL_DEBUG3);
                    #endif
                    }
                else
                    {
                    crate->RestartIdleTimer();
                    bool fatal = (ping_attempts > 0) && (failcount >= ping_attempts);
                    LOGWRITE(Format("Ping: %s failed %d ping(s)%s",
                        ARRAYOFCONST((crate->path, failcount, (fatal) ? ", giving up" : ""))),
                        LOGTASK_CRATE_CTL, (fatal) ? LOGL_ERROR : LOGL_WARNING);
                    if (fatal)
                        {
                        crate->SetState(CRATE_ST_DEAD);
                        return;
                        }
                    }
                break;
            case WAIT_TIMEOUT:
                break;
            default: /* ERROR */
                return;
            }
        }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
__fastcall TIPCrateDataThread::TIPCrateDataThread(TCrateIP *_crate, bool CreateSuspended)
    : TThread(true)
    {
    crate = _crate;
    Priority = tpHigher;
    FreeOnTerminate = false;
    init_complete = CreateEvent(NULL, EVT_MANUAL_RESET, EVT_NONSIGNALED, NULL);
    if (!CreateSuspended) Resume();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
__fastcall TIPCrateDataThread::~TIPCrateDataThread(void)
    {
    CloseHandle(init_complete);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void __fastcall TIPCrateDataThread::Execute(void)
    {
    const TIMEVAL shorter_time = {TIMEOUT_IP_SHORTER / 1000, (TIMEOUT_IP_SHORTER % 1000) * 1000};
    const TIMEVAL longer_time =  {TIMEOUT_IP_LONGER / 1000, (TIMEOUT_IP_LONGER % 1000) * 1000};
    ULONG nonblocking = 1;
    bool keep_trying = true;
    int retry_count = 0;
    TIPCratePingThread *PingThread = NULL;

    crate->special_code = 0;
    crate->total_count_recv = 0;
    crate->total_count_send = 0;

    SetEvent(init_complete); //   

    /*     ,    */
    if (crate->GetState() == CRATE_ST_CONNECTING)
        {
        bool result = crate->InitCrate();
        if (result)
            {
            crate->SetState(CRATE_ST_ONLINE);
            }
        else
            {
            crate->SetState(CRATE_ST_DEAD);
            }
        }

    if (crate->GetState() != CRATE_ST_ONLINE)
        return;

    /*     */
    {
    int attempts = g_Conf.CratePingAttempts();
    if (attempts > 0)
        PingThread = new TIPCratePingThread(crate, false, g_Conf.CratePingTimeout(), attempts);
    }

    Ex(NONFATAL, "Data thread started.", LOGTASK_CRATE_CTL, LOGL_DEBUG1).Dispatch(crate->name);

    while (keep_trying && !Terminated)
        {
        SOCKET sock = INVALID_SOCKET;
        bool connected = false;
        fd_set rset, wset, eset;
        int res;
        DWORD BytesSent = 0;
        DWORD NumSendBytes = 0;
        DWORD skip = 0;
        int RecvShift = 0;
        DWORD BufSize;
        int WaitCount;

        keep_trying = false;
        try
            {
            /*     */
            sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
            if (sock == INVALID_SOCKET)
                throw Ex(FATAL, "Cannot create socket", LOGTASK_CRATE_INIT, LOGL_ERROR, WSAGetLastError());

            /*    */
            BufSize = CRATE_IP_SEND_TCPBUF_SIZE;
            if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char*)&BufSize, sizeof(BufSize)) == SOCKET_ERROR)
                throw Ex(FATAL, "Cannot set socket send buffer", LOGTASK_CRATE_INIT, LOGL_ERROR, WSAGetLastError());
            BufSize = CRATE_IP_RECV_TCPBUF_SIZE;
            if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF, (char*)&BufSize, sizeof(BufSize)) == SOCKET_ERROR)
                throw Ex(FATAL, "Cannot set socket receive buffer", LOGTASK_CRATE_INIT, LOGL_ERROR, WSAGetLastError());

            /*    non-blocking mode */
            if (ioctlsocket(sock, FIONBIO, &nonblocking) == SOCKET_ERROR)
                throw Ex(FATAL, "Cannot set socket to non-blocking mode", LOGTASK_CRATE_INIT, LOGL_ERROR, WSAGetLastError());

            Ex(NONFATAL, "Opening data connection", LOGTASK_CRATE_INIT, LOGL_VERBOSE).Dispatch(crate->name);

            /*   */
            WaitCount = (g_Conf.CrateConnectTimeout() + TIMEOUT_IP_LONGER - 1) / TIMEOUT_IP_LONGER;
            res = crate->ltr_ip->ConnectDataSocket(sock);
            if (res == SOCKET_ERROR)
                {
                int error = WSAGetLastError();
                if (error != WSAEWOULDBLOCK)
                    throw Ex(FATAL, "Connection failed", LOGTASK_CRATE_INIT, LOGL_ERROR, error);
                }

            while (!connected && (WaitCount-- > 0) && !Terminated)
                {
                FD_ZERO(&wset);
                FD_SET(sock, &wset);
                FD_ZERO(&eset);
                FD_SET(sock, &eset);
                res = select(sock + 1, NULL, &wset, &eset, &longer_time);
                if (res == SOCKET_ERROR)
                    throw Ex(FATAL, "Error waiting for connection", LOGTASK_CRATE_INIT, LOGL_ERROR, WSAGetLastError());
                if (FD_ISSET(sock, &eset)) /* connection failed */
                    {
                    DWORD error;
                    int len = sizeof(error);
                    if (getsockopt(sock, SOL_SOCKET, SO_ERROR, (char*)&error, &len) == SOCKET_ERROR)
                        error = WSAEHOSTUNREACH;
                    throw Ex("Connection failed", LOGTASK_CRATE_INIT, LOGL_ERROR, error);
                    }
                if (FD_ISSET(sock, &wset)) /* connection established */
                    connected = true;
                } /* while */

            if (!connected && !Terminated)
                {
                /*  */
                throw Ex(NONFATAL, "Connection timed out", LOGTASK_CRATE_INIT, LOGL_ERROR);
                }

            /*   */
            Ex(NONFATAL, "Data connection established", LOGTASK_CRATE_INIT, LOGL_VERBOSE).Dispatch(crate->name);

            retry_count = 0; /*    */

            while (!Terminated)
                {
                if (BytesSent >= NumSendBytes)
                    {
                    crate->total_count_send += NumSendBytes / sizeof(DWORD);
                    BytesSent = 0;
                    NumSendBytes = sizeof(DWORD) *
                        crate->PrepareDataForCrate((DWORD*)SendBuf, sizeof(SendBuf) / sizeof(DWORD),
                            skip);
                    }
                FD_ZERO(&rset);
                FD_SET(sock, &rset);
                FD_ZERO(&wset);
                if (BytesSent < NumSendBytes)
                    {
                    FD_SET(sock, &wset);
                    res = select(sock + 1, &rset, &wset, NULL, &longer_time);
                    }
                else
                    {
                    res = select(sock + 1, &rset, NULL, NULL, &shorter_time);
                    }
                if (res == SOCKET_ERROR)
                    throw Ex("Error in select()", LOGTASK_CRATE_READ, LOGL_ERROR, WSAGetLastError());

                /* ----  ---- */
                if (FD_ISSET(sock, &rset))
                    {
                    res = recv(sock, RecvBuf + RecvShift, sizeof(RecvBuf) - RecvShift, 0);
                    if (res == SOCKET_ERROR)
                        {
                        int error = WSAGetLastError();
                        if (error != WSAEWOULDBLOCK)
                            throw Ex("Error reading socket", LOGTASK_CRATE_READ, LOGL_ERROR, error);
                        }
                    else if (res == 0)
                        { /*    */
                        throw Ex(NONFATAL, "Connection closed by remote host", LOGTASK_CRATE_READ, LOGL_ERROR);
                        }
                    else
                        { /*  res  */
                        //crate->RestartIdleTimer();
                        int datasize = RecvShift + res;
                        RecvShift = datasize % sizeof(DWORD); /*   */
                        if (datasize >= sizeof(DWORD))
                            {
                            if (!crate->ParseDataFromCrate((DWORD*)RecvBuf, datasize / sizeof(DWORD)))
                                throw Ex(NONFATAL, "Error in crate data", LOGTASK_CRATE_READ, LOGL_ERROR);
                            /*     DWORD,     */
                            for (int i = RecvShift - 1; i >= 0; i--)
                                {
                                RecvBuf[i] = RecvBuf[datasize - RecvShift + i];
                                }
                            }
                        }
                    } /* if can read */

                /* ----  ----*/
                if (FD_ISSET(sock, &wset) && (BytesSent < NumSendBytes))
                    {
                    res = send(sock, SendBuf + BytesSent, NumSendBytes - BytesSent, 0);
                    if (res == SOCKET_ERROR)
                        {
                        int error = WSAGetLastError();
                        if (error != WSAEWOULDBLOCK)
                            throw Ex(NONFATAL, "Error writing socket", LOGTASK_CRATE_WRITE, LOGL_ERROR, error);
                        }
                    else
                        { /*  res  */
                        BytesSent += res;
                        }
                    } /* if can write */
                } /* while !Terminated */
            } /* try */

        catch (Ex ex)
            {
            ex.Dispatch(crate->name);
            keep_trying = !ex.IsFatal();
            }

        /*   */
        if (sock != INVALID_SOCKET)
            {
            if (connected)
                shutdown(sock, SD_BOTH);
            closesocket(sock);
            }

        /*   */
        if (++retry_count >= g_Conf.CrateConnectAttempts())
            keep_trying = false;
        /*     */
        if (keep_trying)
            {
            WaitCount = (g_Conf.CrateReconnectDelay() + TIMEOUT_IP_LONGER - 1) / TIMEOUT_IP_LONGER;
            while ((WaitCount-- > 0) && !Terminated)
                {
                Sleep(TIMEOUT_IP_LONGER);
                }
            }
        else
            { /*         */
            crate->SetState(CRATE_ST_DEAD);
            }
        } /* while (keep_trying && !Terminated) */

    Ex(NONFATAL, "Data thread shutting down", LOGTASK_CRATE_CTL, LOGL_DEBUG2).Dispatch(crate->name);

    if (PingThread)
        {
        PingThread->TerminateByEvent();
        PingThread->WaitFor();
        delete PingThread;
        PingThread = NULL;
        }

    Ex(NONFATAL, "Data thread shut down.", LOGTASK_CRATE_CTL, LOGL_DEBUG1).Dispatch(crate->name);
    }
/*------------------------------------------------------------------------------------------------*/
