#include <vcl.h>
#include <stdio.h>
#pragma hdrstop
#include "LtrCrate.h"
#include "LtrServiceCrates.h"
#include "usb_util.h"
#pragma package(smart_init)

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
#define UDP_SIG_PREFIX          "LTR"
#define UDP_SEND_INTERVAL       2000
#define UDP_MAXERRMSG           10
#define UDP_RCVBUFSIZE          131072

//---------------------------------------------------------------------------
TLTRAutoDiscovery::TLTRAutoDiscovery()
    : working(false), udp_socket(INVALID_SOCKET)
    {
    //       
    send_timer.interval.QuadPart = -10000LL * UDP_SEND_INTERVAL;

    // 
    request.buf.buf = signature;
    request.buf.len = sizeof(signature);
    ZeroMemory(&request.overlap, sizeof(request.overlap));
    reply.buf.buf = recv_buf;
    reply.buf.len = sizeof(recv_buf);
    ZeroMemory(&reply.overlap, sizeof(reply.overlap));

    //   (broadcast)
    ZeroMemory(&bcast_addr, sizeof(bcast_addr));
    bcast_addr.sin_family = AF_INET;
    bcast_addr.sin_port = htons(LTR_IP_IDENTIFY_PORT);
    bcast_addr.sin_addr.s_addr = INADDR_BROADCAST;

    //   
    ZeroMemory(&peer_addr, sizeof(peer_addr));
    peer_addr.sa_family = AF_INET;

    Randomize();
    }
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
TLTRAutoDiscovery::~TLTRAutoDiscovery()
    {
    Stop();
    }
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
void TLTRAutoDiscovery::Start()
    {
    if (working) return; //  

    request.failcount = reply.failcount = 0;

    //    
    strncpy(signature, UDP_SIG_PREFIX, sizeof(signature) - 1);
    for (int i = strlen(UDP_SIG_PREFIX) + 1; i < sizeof(signature); i++)
        {
        signature[i] = Random(256);
        }

    //    
    send_timer.handle = CreateWaitableTimer(NULL, TRUE, NULL);
    request.overlap.hEvent = CreateEvent(NULL, EVT_MANUAL_RESET, EVT_NONSIGNALED, NULL);
    reply.overlap.hEvent = CreateEvent(NULL, EVT_AUTO_RESET, EVT_NONSIGNALED, NULL);

    //  
    udp_socket = WSASocket(AF_INET, SOCK_DGRAM, IPPROTO_UDP, NULL, 0, WSA_FLAG_OVERLAPPED);
    working = (udp_socket != INVALID_SOCKET) &&
              send_timer.handle && reply.overlap.hEvent && request.overlap.hEvent;

    if (working)
        {
        DWORD broadcast = 1;
        if (setsockopt(udp_socket, SOL_SOCKET, SO_BROADCAST,
            (const char*)&broadcast, sizeof(broadcast)) == SOCKET_ERROR)
            {
            LOGWRITE("Cannot enable broadcast on UDP socket", LOGTASK_GLOBAL, LOGL_ERROR);
            working = false;
            }
        }

    if (working)
        {
        DWORD reuse = 1;
        if (setsockopt(udp_socket, SOL_SOCKET, SO_REUSEADDR,
            (const char*)&reuse, sizeof(reuse)) == SOCKET_ERROR)
            {
            LOGWRITE("Cannot enable SO_REUSEADDR on UDP socket", LOGTASK_GLOBAL, LOGL_ERROR);
            working = false;
            }
        }

    if (working)
        {
        DWORD bufsize = UDP_RCVBUFSIZE;
        if (setsockopt(udp_socket, SOL_SOCKET, SO_RCVBUF,
            (const char*)&bufsize, sizeof(bufsize)) == SOCKET_ERROR)
            {
            LOGWRITE("Cannot set UDP socket receive buffer", LOGTASK_GLOBAL, LOGL_ERROR);
            working = false;
            }
        }

    if (working)
        {
        sockaddr_in bind_addr;
        ZeroMemory(&bind_addr, sizeof(bind_addr));
        bind_addr.sin_family = AF_INET;
        bind_addr.sin_port = 0;
        bind_addr.sin_addr.s_addr = INADDR_ANY;
        if (bind(udp_socket, (sockaddr*)&bind_addr, sizeof(bind_addr)) == SOCKET_ERROR)
            {
            LOGWRITE("Cannot bind UDP socket", LOGTASK_GLOBAL, LOGL_ERROR);
            working = false;
            }
        }

    if (working)
        {
        LOGWRITE("UDP auto-discovery mode enabled", LOGTASK_GLOBAL, LOGL_VERBOSE);
        do_Send();
        do_Receive();
        }
    else
        {
        LOGWRITE("Unable to initialize UDP auto-discovery", LOGTASK_GLOBAL, LOGL_ERROR);
        Stop();
        }
    }
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
void TLTRAutoDiscovery::Stop()
    {
    if (working)
        {
        LOGWRITE("UDP auto-discovery mode disabled", LOGTASK_GLOBAL, LOGL_VERBOSE);
        working = false;
        }
    if (udp_socket != INVALID_SOCKET)
        {
        closesocket(udp_socket);
        udp_socket = INVALID_SOCKET;
        }
    if (reply.overlap.hEvent)
        CloseHandle(reply.overlap.hEvent);
    if (request.overlap.hEvent)
        CloseHandle(request.overlap.hEvent);
    if (send_timer.handle)
        CloseHandle(send_timer.handle);
    reply.overlap.hEvent = request.overlap.hEvent = send_timer.handle = NULL;
    }
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
void TLTRAutoDiscovery::do_Send()
    {
    if (!working) return;
    ResetEvent(request.overlap.hEvent);
    if (WSASendTo(udp_socket, &request.buf, 1, NULL, 0,
        (sockaddr*)&bcast_addr, sizeof(bcast_addr), &request.overlap, NULL) == SOCKET_ERROR)
        {
        int err = WSAGetLastError();
        if (err != WSA_IO_PENDING)
            {
            if (request.failcount < UDP_MAXERRMSG)
                {
                LOGWRITE(Format("Cannot send auto-discovery request (error %d)",
                    ARRAYOFCONST((err))), LOGTASK_GLOBAL, LOGL_DEBUG3);
                if (++request.failcount == UDP_MAXERRMSG)
                    LOGWRITE("Further auto-discovery send error messages will be suppressed",
                        LOGTASK_GLOBAL, LOGL_DEBUG3);
                }
            SetEvent(request.overlap.hEvent); //    
            }
        }
    //   (    )
    SetWaitableTimer(send_timer.handle, &send_timer.interval, 0, NULL, NULL, FALSE);
    }
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
void TLTRAutoDiscovery::do_Receive()
    {
    if (!working) return;
    peer_addr_size = sizeof(peer_addr);
    reply.flags = 0;
    if (WSARecvFrom(udp_socket, &reply.buf, 1, NULL, &reply.flags,
        &peer_addr, &peer_addr_size, &reply.overlap, NULL) == SOCKET_ERROR)
        {
        int err = WSAGetLastError();
        if (err != WSA_IO_PENDING)
            {
            LOGWRITE(Format("UDP receive failed (error %d), auto-discovery will not work",
                ARRAYOFCONST((err))), LOGTASK_GLOBAL, LOGL_ERROR);
            //   auto-discovery 
            Stop();
            }
        }
    //   ,   reply.overlap.hEvent
    }
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
bool TLTRAutoDiscovery::Worker(DWORD *pPeerIP)
    {
    HANDLE wait_list[2];
    DWORD peer_ip = 0;

    if (!working)
        {
        if (pPeerIP) *pPeerIP = 0;
        return false;
        }

    //        ,   
    wait_list[0] = request.overlap.hEvent;
    wait_list[1] = send_timer.handle;
    if (WaitForMultipleObjects(2, wait_list, /*all*/ TRUE, 0) == WAIT_OBJECT_0)
        {
        do_Send();
        }

    // ,      - 
    if (WaitForSingleObject(reply.overlap.hEvent, 0) == WAIT_OBJECT_0)
        {
        DWORD recv_len;
        if (WSAGetOverlappedResult(udp_socket, &reply.overlap, &recv_len, FALSE, &reply.flags))
            { //  
            DWORD ip = (peer_addr.sa_family == AF_INET) ?
                ((sockaddr_in*)&peer_addr)->sin_addr.s_addr : 0;

            if ((recv_len == request.buf.len)
                && (!memcmp(request.buf.buf, reply.buf.buf, request.buf.len)))
                {
                // LOGWRITE("Auto-discovery reply received from host " +
                //     IPToStr(peer_addr.sin_addr.s_addr), LOGTASK_GLOBAL, LOGL_DEBUG3);
                peer_ip = ip;
                }
            else
                {
                LOGWRITE("Invalid auto-discovery reply from host" +
                    IPToStr(ip), LOGTASK_GLOBAL, LOGL_WARNING);
                }
            }
        else
            { //    
            int err = WSAGetLastError();
            if (reply.failcount < UDP_MAXERRMSG)
                {
                LOGWRITE(Format("Auto-discovery receive error (%d)",
                    ARRAYOFCONST((err))), LOGTASK_GLOBAL, LOGL_DEBUG3);
                if (++reply.failcount == UDP_MAXERRMSG)
                    LOGWRITE("Further auto-discovery receive error messages will be suppressed",
                        LOGTASK_GLOBAL, LOGL_DEBUG3);
                }
            }
        do_Receive();
        }

    if (pPeerIP) *pPeerIP = peer_ip;
    return (peer_ip != 0);
    }
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
__fastcall TLTRServiceCrate::TLTRServiceCrate(bool CreateSuspended, MessageQueue<MsgDisplay>& mqDisplay)
    : TThread(true),
    init_ok(false),
    refMQ_Display(mqDisplay)
    {
    Priority = tpNormal;
    FreeOnTerminate = false;

    init_complete = CreateEvent(NULL, EVT_MANUAL_RESET, EVT_NONSIGNALED, NULL);
    if (!CreateSuspended) Resume();
    }
//---------------------------------------------------------------------------
__fastcall TLTRServiceCrate::~TLTRServiceCrate(void)
    {
    CloseHandle(init_complete);
    }
//---------------------------------------------------------------------------
//  
//---------------------------------------------------------------------------
/*------------------------------------------------------------------------------------------------*/
void __fastcall TLTRServiceCrate::StopServiceForDeadCrates(void)
    {
    if (WaitForSingleObject(crate_list_access, INFINITE) != WAIT_OBJECT_0)
        return;
    //***     
    for (int i = 0; i < CRATE_MAX; i++)
        {
        if (crate_list[i]->GetState() == CRATE_ST_DEAD)
            {
            LOGWRITE(Format("%s: crate stopped",
                ARRAYOFCONST((crate_list[i]->name))),
                LOGTASK_CRATE_CTL, LOGL_INFO);
            delete crate_list[i];
            crate_list[i] = NULL;
            }
        }
    ReleaseMutex(crate_list_access);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void __fastcall TLTRServiceCrate::ConnectUSBCrate(DWORD DevInst, const char* szPath)
    {
    AnsiString strPath, strHubID, errstr;
    const char* device_path = szPath;
    bool ok;
    t_USB_ConnInfo conn_info;

    if (!device_path)
        { /*     */
        ok = USB_GetDeviceInfoByInstance(DevInst, &strPath, &conn_info, errstr);
        device_path = strPath.c_str();
        }
    else
        { /*  ,    */
        ok = USB_GetDeviceInfoByInstance(DevInst, NULL, &conn_info, errstr);
        }

    if (!ok)
        {
        LOGWRITE(Format("Cannot get device information (devinst = %0.8X): %s",
            ARRAYOFCONST((DevInst, errstr))),
            LOGTASK_CRATE_INIT, LOGL_ERROR);
        return;
        }
    else if (!errstr.IsEmpty())
        {
        LOGWRITE(Format("Cannot get device speed (devinst = %0.8X): %s",
            ARRAYOFCONST((DevInst, errstr))),
            LOGTASK_CRATE_INIT, LOGL_WARNING);
        }

    if (g_Log.CheckLevel(LOGL_DEBUG1))
        {
        LOGWRITE(Format("Connecting USB crate: devinst = %0.8X, path = %s",
            ARRAYOFCONST((DevInst, device_path))),
            LOGTASK_CRATE_INIT, LOGL_INFO);
        }
    else
        {
        LOGWRITE("Connecting USB crate...", LOGTASK_CRATE_INIT, LOGL_INFO);
        }

    if (WaitForSingleObject(crate_list_access, INFINITE) != WAIT_OBJECT_0)
        return;

    //   
    int FreeSlot = -1;
    for (int i = 0; (FreeSlot < 0) && (i < CRATE_MAX); i++)
        {
        if (!crate_list[i])
            FreeSlot = i;
        }

    if (FreeSlot >= 0)
        {
        TCrate *crate = new TCrateUSB(MQ_Control, refMQ_Display);
        if (crate)
            {
            crate->usb_devinst = DevInst;
            crate->usb_info = conn_info;
            STRNCPYZ(crate->path, device_path);
            crate_list[FreeSlot] = crate;
            /*    */
            if (!crate->Startup(NULL, true))
                {
                LOGWRITE(Format("Cannot initialize crate %s", ARRAYOFCONST((device_path))),
                    LOGTASK_CRATE_INIT, LOGL_ERROR);
                }
            }
        else
            {
            LOGWRITE(Format("Cannot create instance for crate %s", ARRAYOFCONST((device_path))),
                LOGTASK_CRATE_INIT, LOGL_ERROR);
            }
        }
    else /* no free slot */
        {
        LOGWRITE(Format("No free slot for crate %s (maximum %d reached)",
            ARRAYOFCONST((device_path, (int)CRATE_MAX))),
            LOGTASK_CRATE_INIT, LOGL_ERROR);
        }
    ReleaseMutex(crate_list_access);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void __fastcall TLTRServiceCrate::DisconnectUSBCrate(DWORD DevInst)
    {
    LOGWRITE(Format("Disconnecting USB crate: devinst = %0.8X",
        ARRAYOFCONST((DevInst))),
        LOGTASK_CRATE_INIT, LOGL_DEBUG3);

    if (WaitForSingleObject(crate_list_access, INFINITE) != WAIT_OBJECT_0)
        return;

    for (int ci = 0; ci < CRATE_MAX; ci++)
        {
        if (crate_list[ci]
            && (crate_list[ci]->GetState() != CRATE_ST_DEAD)
            && (crate_list[ci]->GetConnType() == CRATE_IFACE_USB)
//            && !stricmp(crate_list[ci]->path, device_path))
            && (crate_list[ci]->usb_devinst == DevInst))
            {
            crate_list[ci]->SetState(CRATE_ST_DEAD);
            }
        }

    ReleaseMutex(crate_list_access);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
/*     USB
 * USB_DETECT_ADVANCED = 1  ==>      ( 1   )
 * USB_DETECT_ADVANCED = 0  ==>      (  devnodes_changed) */
void __fastcall TLTRServiceCrate::ScanUSBCrates(void)
    {
#if !USB_DETECT_ADVANCED
#define USBCRATEINFO_MAX 32
    struct t_CrateInfo
        {
        AnsiString path;
        DWORD devinst;
        bool connected;
        }
        USBCrates[USBCRATEINFO_MAX];
    int USBCrateCount = 0;
#endif
    SP_DEVICE_INTERFACE_DATA IfaceData;
    AnsiString path, desc, errstr;
    DWORD DevInst;
    DWORD MemberIdx = 0;

    LOGWRITE("Scanning USB crates", LOGTASK_CRATE_CTL, LOGL_VERBOSE);

    /*     USB  */
    HDEVINFO hList =
        SetupDiGetClassDevs(&USB_DEVICE_CLSID, NULL, NULL, DIGCF_PRESENT | DIGCF_DEVICEINTERFACE);
    if (INVALID_HANDLE_VALUE == hList)
        {
        int err = GetLastError();
        LOGWRITE(Format("Cannot enumerate USB (SetupDiGetClassDevs() error %d)",
            ARRAYOFCONST((err))), LOGTASK_CRATE_CTL, LOGL_VERBOSE);
        return;
        }

    /*    */
#if !USB_DETECT_ADVANCED
    while (USBCrateCount < USBCRATEINFO_MAX)
#else
    while (true)
#endif
        {
        ZeroMemory(&IfaceData, sizeof(IfaceData));
        IfaceData.cbSize = sizeof(IfaceData);
        if (SetupDiEnumDeviceInterfaces(hList, NULL, &USB_DEVICE_CLSID, MemberIdx++, &IfaceData))
            {
            if (!(IfaceData.Flags & SPINT_ACTIVE))
                continue;

            if (!USB_GetDeviceInfoByIface(hList, &IfaceData, &path, &desc, &DevInst, errstr))
                {
                LOGWRITE("Cannot process USB device: " + errstr,
                    LOGTASK_GLOBAL, LOGL_DEBUG1);
                continue;
                }

            /* ,   ,   VID/PID */
            if (USB_IsKnownDeviceName(path))
                {
                if (g_Log.CheckLevel(LOGL_DEBUG1))
                    {
                    LOGWRITE(Format("USB crate found: \"%s\" at %s",
                        ARRAYOFCONST((desc, path))),
                        LOGTASK_CRATE_CTL, LOGL_INFO);
                    }
                else
                    {
                    LOGWRITE(Format("USB crate found: \"%s\"", ARRAYOFCONST((desc))),
                        LOGTASK_CRATE_CTL, LOGL_INFO);
                    }
#if !USB_DETECT_ADVANCED
                USBCrates[USBCrateCount].path = path;
                USBCrates[USBCrateCount].devinst = DevInst;
                USBCrates[USBCrateCount].connected = false;
                if (++USBCrateCount >= USBCRATEINFO_MAX)
                    LOGWRITE("Warning: no more USB crates can be processed", LOGTASK_CRATE_CTL, LOGL_DEBUG1);
#else
                ConnectUSBCrate(DevInst, path.c_str());
#endif
                }
            else
                {
                LOGWRITE(Format("USB device found, unsupported: \"%s\" at %s",
                    ARRAYOFCONST((desc, path))),
                    LOGTASK_CRATE_CTL, LOGL_DEBUG3);
                }
            }
        else /* SetupDiEnumDeviceInterfaces returned false */
            {
            int err = GetLastError();
            if (err != ERROR_NO_MORE_ITEMS)
                {
                LOGWRITE(Format("Cannot enumerate USB (SetupDiEnumDeviceInterfaces() error %d)",
                    ARRAYOFCONST((err))), LOGTASK_CRATE_CTL, LOGL_ERROR);
                }
            break; /*     */
            }
        } /* while */
    SetupDiDestroyDeviceInfoList(hList);


#if !USB_DETECT_ADVANCED
    /*   USBCrates[]       (   USB) */

    if (USBCrateCount > 0)
        {
        LOGWRITE(Format("Found %d USB crate(s)", ARRAYOFCONST((USBCrateCount))),
            LOGTASK_CRATE_CTL, LOGL_DEBUG1);
        }
    else
        {
        LOGWRITE("No USB crates found", LOGTASK_CRATE_CTL, LOGL_DEBUG1);
        }

    /*       ,      */
    if (WaitForSingleObject(crate_list_access, INFINITE) == WAIT_OBJECT_0)
        {
        /*   */
        for (int ci = 0; ci < CRATE_MAX; ci++)
            {
            if (crate_list[ci]
                && (crate_list[ci]->GetState() != CRATE_ST_DEAD)
                && (crate_list[ci]->GetConnType() == CRATE_IFACE_USB))
                {
                bool still_connected = false;
                for (int ai = 0; ai < USBCrateCount; ai++)
                    {
                    if (USBCrates[ai].devinst == crate_list[ci]->usb_devinst)
//                    if (!stricmp(USBCrates[ai].path.c_str(), crate_list[ci]->path))
                        {
                        USBCrates[ai].connected = true; /* ,     */
                        still_connected = true;
                        break;
                        }
                    }
                if (!still_connected)
                    crate_list[ci]->SetState(CRATE_ST_DEAD);
                }
            }
        /*   */
        for (int ai = 0; ai < USBCrateCount; ai++)
            {
            if (!USBCrates[ai].connected)
                ConnectUSBCrate(USBCrates[ai].devinst, USBCrates[ai].path.c_str());
            } /* for ai */

        ReleaseMutex(crate_list_access);
        }
#endif
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void __fastcall TLTRServiceCrate::ConnectCrateByIP(DWORD IP)
    {
    bool AlreadyExists = false;
    int FreeSlot = -1;

    TAddrListNode addr;
    if (!addr.SetAddr(IP)) return;

    /* ,       */
    if (WaitForSingleObject(crate_list_access, INFINITE) != WAIT_OBJECT_0)
        return;
    for (int i = 0; i < CRATE_MAX; i++)
        {
        if (!crate_list[i])
            {
            if (FreeSlot < 0)
                FreeSlot = i;
            }
        else if ((crate_list[i]->GetState() != CRATE_ST_DEAD)
                && (crate_list[i]->ip_addr == addr.dwIP))
            {
            AlreadyExists = true;
            break;
            }
        }
    /*       ,    */
    if (!AlreadyExists)
        {
        AnsiString ip_str = IPToStr(addr.dwIP);
        if (FreeSlot >= 0)
            {
            TCrate *crate = new TCrateIP(MQ_Control, refMQ_Display, g_Conf.CrateIdleTimeout());
            if (crate)
                {
                crate->ip_addr = addr.dwIP;
                STRNCPYZ(crate->path, ip_str.c_str());
                crate_list[FreeSlot] = crate;
                /*     () */
                if (!crate->Startup(NULL, true))
                    {
                    LOGWRITE("Cannot initialize crate " + ip_str,
                        LOGTASK_CRATE_INIT, LOGL_ERROR);
                    }
                }
            }
        else /* no free slot */
            {
            LOGWRITE(Format("No free slot for crate %s (maximum %d reached)",
                ARRAYOFCONST((ip_str, (int)CRATE_MAX))),
                LOGTASK_CRATE_INIT, LOGL_ERROR);
            }
        } /* !AlreadyExists */
    ReleaseMutex(crate_list_access);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void __fastcall TLTRServiceCrate::DisconnectCrateByIP(DWORD IP)
    {
    /*      */
    if (WaitForSingleObject(crate_list_access, INFINITE) != WAIT_OBJECT_0)
        return;
    for (int i = 0; i < CRATE_MAX; i++)
        {
        if (crate_list[i] && (crate_list[i]->GetState() != CRATE_ST_DEAD)
            && (crate_list[i]->ip_addr == IP))
            {
            crate_list[i]->SetState(CRATE_ST_DEAD);
            break;
            }
        }
    ReleaseMutex(crate_list_access);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void __fastcall TLTRServiceCrate::ConnectAllIPCratesByType(int IPListType)
    {
    /*   ,   AUTOCONNECT */
    TAddrList *IPList;
    TAddrListNode *addr;

    IPList = IPCrateList.ExportList(IPListType);
    if (!IPList)
        return;
    addr = IPList->GetFirstNode();
    while (addr)
        {
        if (addr->dwFlags & CRATE_IP_FLAG_AUTOCONNECT)
            {
            switch (addr->iStatus)
                {
                case IP_STATUS_OFFLINE:
                    /*  --  */
                    ConnectCrateByIP(addr->dwIP);
                    break;
                case IP_STATUS_CONNECTING:
                    /*  --     */
                    DisconnectCrateByIP(addr->dwIP);
                    StopServiceForDeadCrates();
                    ConnectCrateByIP(addr->dwIP);
                    break;
                case IP_STATUS_ONLINE:
                    /*   --   */
                    break;
                }
            }
        addr = addr->next;
        } /* while (addr) */
    delete IPList;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void __fastcall TLTRServiceCrate::DisconnectAllIPCratesByType(int IPListType)
    {
    TAddrList *IPList;
    TAddrListNode *addr;

    IPList = IPCrateList.ExportList(IPListType);
    if (!IPList)
        return;
    addr = IPList->GetFirstNode();
    while (addr)
        {
        switch (addr->iStatus)
            {
            case IP_STATUS_CONNECTING:
            case IP_STATUS_ONLINE:
                /*   */
                DisconnectCrateByIP(addr->dwIP);
                break;
            }
        addr = addr->next;
        } /* while (addr) */
    delete IPList;
    }
/*------------------------------------------------------------------------------------------------*/

//-----------------------------------------------------------------------------
void __fastcall TLTRServiceCrate::SuspendCrates()
{
    LOGWRITE("LTRServiceCrate going to suspend.", LOGTASK_CRATE_CTL, LOGL_VERBOSE);
    if(WaitForSingleObject(crate_list_access, INFINITE)==WAIT_OBJECT_0)
    {
        for(int i=0; i<CRATE_MAX; i++)
        {
            if(crate_list[i]!=NULL) crate_list[i]->SuspendCrate();
        }
        ReleaseMutex(crate_list_access);
    }
}
//-----------------------------------------------------------------------------
void __fastcall TLTRServiceCrate::ResumeCrates()
{
    LOGWRITE("LTRServiceCrate resuming from suspend.", LOGTASK_CRATE_CTL, LOGL_VERBOSE);
    if(WaitForSingleObject(crate_list_access, INFINITE)==WAIT_OBJECT_0)
    {
        for(int i=0; i<CRATE_MAX; i++)
        {
            if(crate_list[i]!=NULL) crate_list[i]->ResumeCrate();
        }
        ReleaseMutex(crate_list_access);
    }
}

//-----------------------------------------------------------------------------
void __fastcall TLTRServiceCrate::Execute()
    {
    DWORD found_ip;
    MsgControl CurrMessage;

    LOGWRITE("LTRServiceCrate thread started.", LOGTASK_CRATE_CTL, LOGL_VERBOSE);

    //  
    MQ_Control.put(LTR_SRV_rescan_usb);
    MQ_Control.put(LTR_SRV_server_priority_changed);
    MQ_Control.put(LTR_SRV_autoconnect_ip);
    refMQ_Display.put(LTR_DISP_crate_list_updated);
    refMQ_Display.put(LTR_DISP_ip_list_updated);

    //  auto-discovery
    if (g_Conf.EnableAutoDiscovery())
        AutoDiscovery.Start();

    //         
    ltrserviceclient = new TLTRServiceClient(false, this);
    if (ltrserviceclient)
        {
        if (WaitForSingleObject(ltrserviceclient->init_complete, 5000) == WAIT_OBJECT_0)
            {
            if (ltrserviceclient->init_ok)
                {
                //   
                init_ok = true;
                SetEvent(init_complete);
                }
            else
                {
                LOGWRITE("Initialization of LTRServiceClient thread failed.", LOGTASK_GLOBAL, LOGL_CRITICAL);
                StopThread(ltrserviceclient);
                }
            }
        else
            {
            LOGWRITE("Cannot start LTRServiceClient thread.", LOGTASK_GLOBAL, LOGL_CRITICAL);
            TerminateThread((HANDLE)ltrserviceclient->Handle, 0);
            }
        }
    else
        {
        LOGWRITE("Cannot create LTRServiceClient thread.", LOGTASK_GLOBAL, LOGL_CRITICAL);
        }

    while (!Terminated && init_ok)
        {
        if (MQ_Control.get(&CurrMessage, 200))
            {
            switch (CurrMessage.dwMessage)
                {
                case LTR_SRV_insert_usb:
                    ConnectUSBCrate(CurrMessage.dwParam);
                    break;
                case LTR_SRV_remove_usb:
                    DisconnectUSBCrate(CurrMessage.dwParam);
                    break;
                case LTR_SRV_rescan_usb:
                    ScanUSBCrates();
                    break;
                case LTR_SRV_autodiscovery_changed:
                    if (g_Conf.EnableAutoDiscovery())
                        {
                        AutoDiscovery.Start();
                        }
                    else
                        {
                        AutoDiscovery.Stop();
                        }
                    break;
                case LTR_SRV_autoconnect_ip:
                    ConnectAllIPCratesByType(IPLIST_STATIC);
                    ConnectAllIPCratesByType(IPLIST_DYNAMIC);
                    break;
                case LTR_SRV_disconnect_all_ip:
                    DisconnectAllIPCratesByType(IPLIST_STATIC);
                    DisconnectAllIPCratesByType(IPLIST_DYNAMIC);
                    break;
                case LTR_SRV_server_priority_changed:
                    if (SetPriorityClass(GetCurrentProcess(), g_Conf.ServerPriority()))
                        {
                        refMQ_Display.put(LTR_DISP_gui_vars_updated);
                        }
                    else
                        {
                        LOGWRITE("Error setting server priority", LOGTASK_CRATE_CTL, LOGL_ERROR);
                        }
                    break;
                case LTR_SRV_suspend:
                    this->SuspendCrates();
                    break;
                case LTR_SRV_resume_suspend:
                    this->ResumeCrates();
                    break;
                case LTR_SRV_bury_the_dead:
                    StopServiceForDeadCrates();
                    break;
                case LTR_SRV_connect_crate:
                    ConnectCrateByIP(CurrMessage.dwParam);
                    break;
                case LTR_SRV_disconnect_crate:
                    DisconnectCrateByIP(CurrMessage.dwParam);
                    break;
                default:
                    LOGWRITE(Format("Unknown Service Crate Incoming Message=%d", ARRAYOFCONST((CurrMessage.dwMessage))),
                        LOGTASK_CRATE_CTL, LOGL_ERROR);
                    break;
                }
            }

        //       (UDP auto-discover)
        if (AutoDiscovery.Worker(&found_ip))
            {
            DWORD flags = (g_Conf.AutoConnectDynamic() ? CRATE_IP_FLAG_AUTOCONNECT : 0);
            if (IPCrateList.AddUnique(IPLIST_DYNAMIC, found_ip, flags))
                {
                LOGWRITE("Found new IP crate: " + IPToStr(found_ip),
                    LOGTASK_CRATE_INIT, LOGL_INFO);
                refMQ_Display.put(LTR_DISP_ip_list_updated);
                if (flags & CRATE_IP_FLAG_AUTOCONNECT)
                    MQ_Control.put(MsgControl(LTR_SRV_connect_crate, found_ip));
                }
            }

        } /* while (!Terminated) */

    /* Terminated */

    AutoDiscovery.Stop();

    /*     */
    SetPriorityClass(GetCurrentProcess(), NORMAL_PRIORITY_CLASS);

    //    ,  
    if (ltrserviceclient)
        {
        ltrserviceclient->Terminate();
        if (ltrserviceclient->server != INVALID_SOCKET)
            { /*  ,  accept() */
            shutdown(ltrserviceclient->server, SD_BOTH);
            closesocket(ltrserviceclient->server);
            ltrserviceclient->server = INVALID_SOCKET;
            }
        StopThread(ltrserviceclient);
        }

    //     ,   
    if (WaitForSingleObject(crate_list_access, INFINITE) == WAIT_OBJECT_0)
        {
        for (int i = 0; i < CRATE_MAX; i++)
            {
            if (crate_list[i])
                {
                LOGWRITE(Format("%s: crate stopped",
                    ARRAYOFCONST((crate_list[i]->name))),
                    LOGTASK_CRATE_CTL, LOGL_INFO);
                delete crate_list[i];
                crate_list[i] = NULL;
                }
            }
        ReleaseMutex(crate_list_access);
        }
    else
        {
        LOGWRITE("Cannot access crate list.", LOGTASK_CRATE_CTL, LOGL_ERROR);
        }
    LOGWRITE("LTRServiceCrate thread shut down.", LOGTASK_CRATE_CTL, LOGL_VERBOSE);
    SetEvent(init_complete);
    }
//---------------------------------------------------------------------------

