#include <windows.h>
#include <cfgmgr32.h>
#include <winioctl.h>
#include "usb_util.h"
#include "ddk/usbioctl.h"
#include "ddk/usb200.h"

/*================================================================================================*/
const CLSID USB_DEVICE_CLSID =
    {
    0xA5DCBF10, 0x6530, 0x11D2, { 0x90, 0x1F, 0x00, 0xC0, 0x4F, 0xB9, 0x51, 0xED }
    };
const CLSID USB_HUB_CLSID =
    {
    0xF18A0E88, 0xC30C, 0x11D0, { 0x88, 0x15, 0x00, 0xA0, 0xC9, 0x06, 0xBE, 0xD8 }
    };

#define PORTINFO_BUFFER_SIZE \
    (sizeof(USB_NODE_CONNECTION_INFORMATION_EX) + sizeof(USB_PIPE_INFO) * 30)

static const AnsiString KnownUSBDevices[] =
    {
    "vid_0471&pid_1010", // LTR010
    "vid_0471&pid_2121", // LTR021
    "vid_0471&pid_3030", // LTR030
    "vid_0471&pid_3131"  // LTR031
    };
/*================================================================================================*/

/*================================================================================================*/
class nocopy
    {
protected:
    nocopy() { }
    //~nocopy() { }
private:
    nocopy(const nocopy&);
    const nocopy& operator= (const nocopy&);
    };

//  : ,      
template<typename T> class auto_buf : nocopy
    {
private:
    T* data_ptr;
    size_t data_size;
    void close()
        { if (data_ptr) free(data_ptr); }
public:
    auto_buf() : data_size(0), data_ptr(NULL)
        { }
    size_t alloc(size_t nbytes)
        {
        close();
        data_ptr = (nbytes) ? (T*)malloc(nbytes) : NULL;
        data_size = (data_ptr) ? nbytes : 0;
        return data_size;
        }
    ~auto_buf()                 { close(); }
    T* ptr() const              { return data_ptr; }
    size_t size() const         { return data_size; }
    };

//  :       
class auto_devlist : nocopy
    {
private:
    HDEVINFO handle;
    void close()
        { if (INVALID_HANDLE_VALUE != handle) SetupDiDestroyDeviceInfoList(handle); }
public:
    auto_devlist() : handle(INVALID_HANDLE_VALUE)
        { }
    bool open(HDEVINFO new_handle)
        {
        close();
        return (INVALID_HANDLE_VALUE != (handle = new_handle));
        }
    ~auto_devlist()             { close(); }
    HDEVINFO Handle() const     { return handle; }
    };
/*================================================================================================*/

/*================================================================================================*/
/*------------------------------------------------------------------------------------------------*/
static bool f_GetDevIdByInstance            //   device instance ID   DEVINST
    (
    DWORD hDevInst,                         // ()    
    auto_buf<char>& buf,                    // () ,    
    AnsiString& strErrorMsg                 // ()   ,    false
    )
    {
    DWORD id_size;
    //    (  NULL-)
    DWORD cm_res = CM_Get_Device_ID_Size(&id_size, hDevInst, 0);
    if (CR_SUCCESS != cm_res)
        {
        strErrorMsg = "CM_Get_Device_ID_Size() error " + IntToStr(cm_res);
        return false;
        }
    if (!id_size)
        {
        strErrorMsg = "Device ID not found";
        return false;
        }
    id_size++; //  NULL-
    //  
    if (!buf.alloc(id_size))
        {
        strErrorMsg = "Failed to allocate " + IntToStr(id_size) + " bytes for DevId";
        return false;
        }
    //  
    cm_res = CM_Get_Device_ID(hDevInst, buf.ptr(), buf.size(), 0);
    if (CR_SUCCESS != cm_res)
        {
        strErrorMsg = "CM_Get_Device_ID() error " + IntToStr(cm_res);
        return false;
        }
    return true;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static bool f_GetDeviceByIface              //  ,  DevicePath
    (
    HDEVINFO hList,                         // ()    
    PSP_DEVICE_INTERFACE_DATA pIfaceData,   // ()     hList
    auto_buf
        <SP_DEVICE_INTERFACE_DETAIL_DATA>*
        pbuf,                               // () ,  
    PSP_DEVINFO_DATA pDevInfoData,          // ()   ( NULL)
    AnsiString& strErrorMsg                 // ()   ,    false
    )
    {
    int err = 0;
    DWORD data_size;

    //    interface detail
    if (SetupDiGetDeviceInterfaceDetail(hList, pIfaceData, NULL, 0, &data_size, pDevInfoData)
        || ((err = GetLastError()) != ERROR_INSUFFICIENT_BUFFER))
        {
        strErrorMsg = "SetupDiGetDeviceInterfaceDetail(size) error " + IntToStr(err);
        return false;
        }
    //     (buf == NULL),  
    if (!pbuf)
        return true;

    //  
    if (!pbuf->alloc(data_size))
        {
        strErrorMsg = "Failed to allocate " + IntToStr(data_size) + " bytes for interface detail";
        return false;
        }
    //  
    pbuf->ptr()->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);
    if (!SetupDiGetDeviceInterfaceDetail(hList, pIfaceData, pbuf->ptr(), pbuf->size(), NULL, NULL))
        {
        err = GetLastError();
        strErrorMsg = "SetupDiGetDeviceInterfaceDetail(data) error " + IntToStr(err);
        return false;
        }
    return true;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static bool f_GetDeviceProperty
    (
    HDEVINFO hList,                         // ()    
    PSP_DEVINFO_DATA pDevInfoData,          // ()       hList
    DWORD Property,                         // ()   property (SPDRP_...)
    DWORD RequiredPropType,                 // ()     (REG_...), 0 = 
    auto_buf<char>& buf,                    // () ,    property
    AnsiString& strErrorMsg                 // ()   ,    false
    )
    {
    int err = 0;
    DWORD data_size, data_type;

    //   
    if (SetupDiGetDeviceRegistryProperty(hList, pDevInfoData, Property, &data_type, NULL, 0, &data_size)
        || ((err = GetLastError()) != ERROR_INSUFFICIENT_BUFFER))
        {
        strErrorMsg = "SetupDiGetDeviceRegistryProperty(size) error " + IntToStr(err);
        return false;
        }
    if (RequiredPropType && (data_type != RequiredPropType))
        {
        strErrorMsg = "Property type mismatch (need " +
            IntToStr(RequiredPropType) + ", got " + IntToStr(data_type) + ")";
        return false;
        }
    //  
    if (!buf.alloc(data_size))
        {
        strErrorMsg = "Failed to allocate " + IntToStr(data_size) + " bytes for property";
        return false;
        }
    //  
    if (!SetupDiGetDeviceRegistryProperty(hList, pDevInfoData, Property, NULL, buf.ptr(), buf.size(), NULL))
        {
        err = GetLastError();
        strErrorMsg = "SetupDiGetDeviceRegistryProperty(data) error " + IntToStr(err);
        return false;
        }
    return true;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static HANDLE f_OpenParentHubByInstance
    (
    DWORD DevInst,                          // ()    (DEVINST)
    AnsiString& strErrorMsg                 // ()   ,    false
    )
    {
    DWORD ParentDevInst;
    SP_DEVICE_INTERFACE_DATA IfaceData;
    auto_devlist List;

    //  DEVINST  
    DWORD cm_res = CM_Get_Parent(&ParentDevInst, DevInst, 0);
    if (CR_SUCCESS != cm_res)
        {
        strErrorMsg = "CM_Get_Parent() error " + IntToStr(cm_res);
        return INVALID_HANDLE_VALUE;
        }

    //  Device Instance ID  
    {
    auto_buf<char> DevId;
    if (!f_GetDevIdByInstance(ParentDevInst, DevId, strErrorMsg))
        return INVALID_HANDLE_VALUE;

    //   DevId  
    if (!List.open(SetupDiGetClassDevs(&USB_HUB_CLSID, DevId.ptr(), NULL, DIGCF_DEVICEINTERFACE)))
        {
        strErrorMsg = "Cannot create device info list";
        return INVALID_HANDLE_VALUE;
        }
    } // DevId   

    strErrorMsg = "No interfaces on hub";
    for (int i = 0; ; i++)
        {
        ZeroMemory(&IfaceData, sizeof(IfaceData));
        IfaceData.cbSize = sizeof(IfaceData);
        if (SetupDiEnumDeviceInterfaces(List.Handle(), NULL, &USB_HUB_CLSID, i, &IfaceData))
            {
            //       
            auto_buf<SP_DEVICE_INTERFACE_DETAIL_DATA> detail;
            if (f_GetDeviceByIface(List.Handle(), &IfaceData, &detail, NULL, strErrorMsg))
                {
                HANDLE hHub = CreateFile(detail.ptr()->DevicePath, GENERIC_WRITE, FILE_SHARE_WRITE,
                    NULL, OPEN_EXISTING, 0, NULL);
                if (INVALID_HANDLE_VALUE != hHub)
                    {
                    strErrorMsg = "";
                    return hHub;
                    }
                strErrorMsg = "CreateFile() error " + IntToStr(GetLastError());
                }
            }
        else
            {
            int err = GetLastError();
            if (err != ERROR_NO_MORE_ITEMS)
                {
                strErrorMsg = "SetupDiEnumDeviceInterfaces() error " + IntToStr(err);
                return INVALID_HANDLE_VALUE;
                }
            break; // no more items
            }
        } /* for i */

    // strErrorMsg   
    return INVALID_HANDLE_VALUE;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static int f_GetHubPortCount
    (
    HANDLE hHub
    )
    {
    DWORD XferSize;
    USB_NODE_INFORMATION NodeInfo;
    ZeroMemory(&NodeInfo, sizeof(NodeInfo));
    if (!DeviceIoControl(hHub, IOCTL_USB_GET_NODE_INFORMATION,
        &NodeInfo, sizeof(NodeInfo), &NodeInfo, sizeof(NodeInfo), &XferSize, NULL))
        return -1;
    return NodeInfo.u.HubInformation.HubDescriptor.bNumberOfPorts;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static bool f_GetHubPortInfo
    (
    HANDLE hHub,
    int nPort,
    auto_buf<char>& buf,
    bool& bConnected,
    t_USB_ConnInfo& Info
    )
    {
    DWORD XferSize;
    if (buf.size() < PORTINFO_BUFFER_SIZE)
        {
        if (!buf.alloc(PORTINFO_BUFFER_SIZE))
            return false;
        }
    PUSB_NODE_CONNECTION_INFORMATION_EX p =
        (PUSB_NODE_CONNECTION_INFORMATION_EX)buf.ptr();
    p->ConnectionIndex = nPort;
    if (DeviceIoControl(hHub, IOCTL_USB_GET_NODE_CONNECTION_INFORMATION_EX,
        buf.ptr(), buf.size(), buf.ptr(), buf.size(), &XferSize, NULL))
        {
        bConnected = (DeviceConnected == p->ConnectionStatus);
        Info.usb_version = p->DeviceDescriptor.bcdUSB;
        switch (p->Speed)
            {
            case UsbLowSpeed:   Info.speed = USB_Speed_Low;  break;
            case UsbFullSpeed:  Info.speed = USB_Speed_Full; break;
            case UsbHighSpeed:
            default:            Info.speed = USB_Speed_High; break;
            }
        }
    else
        {
        PUSB_NODE_CONNECTION_INFORMATION p =
            (PUSB_NODE_CONNECTION_INFORMATION)buf.ptr();
        p->ConnectionIndex = nPort;
        if (DeviceIoControl(hHub, IOCTL_USB_GET_NODE_CONNECTION_INFORMATION,
            buf.ptr(), buf.size(), buf.ptr(), buf.size(), &XferSize, NULL))
            {
            bConnected = (DeviceConnected == p->ConnectionStatus);
            Info.usb_version = p->DeviceDescriptor.bcdUSB;
            Info.speed = (p->LowSpeed) ? USB_Speed_Low : USB_Speed_Full;
            }
        else
            {
            return false;
            }
        }
    return true;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static bool f_GetHubPortDriverKey
    (
    HANDLE hHub,
    int nPort,
    auto_buf<char>& buf
    )
    {
    DWORD XferSize;
    USB_NODE_CONNECTION_DRIVERKEY_NAME size_req;
    size_req.ConnectionIndex = nPort;
    size_req.ActualLength = 0;
    if (!DeviceIoControl(hHub, IOCTL_USB_GET_NODE_CONNECTION_DRIVERKEY_NAME,
            &size_req, sizeof(size_req), &size_req, sizeof(size_req), &XferSize, NULL))
        return false;

    auto_buf<USB_NODE_CONNECTION_DRIVERKEY_NAME> DriverKey;
    if (!DriverKey.alloc(size_req.ActualLength))
        return false;
    DriverKey.ptr()->ConnectionIndex = nPort;
    if (!DeviceIoControl(hHub, IOCTL_USB_GET_NODE_CONNECTION_DRIVERKEY_NAME,
            DriverKey.ptr(), DriverKey.size(), DriverKey.ptr(), DriverKey.size(),
            &XferSize, NULL))
        return false;

    // DriverKey.ptr()->DriverKeyName    UNICODE
    size_t ansi_size = 
        WideCharToMultiByte(CP_ACP, 0, DriverKey.ptr()->DriverKeyName, -1, NULL, 0, NULL, NULL);
    if (!ansi_size || !buf.alloc(ansi_size))
        return false;
    return 
        (WideCharToMultiByte(CP_ACP, 0, DriverKey.ptr()->DriverKeyName, -1,
            buf.ptr(), buf.size(), NULL, NULL) > 0);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static bool f_GetDeviceConnInfo
    (
    HDEVINFO hList,                         // ()    
    PSP_DEVINFO_DATA pDevInfoData,          // ()       hList
    t_USB_ConnInfo& Info,                   // ()    
    AnsiString& strErrorMsg                 // ()   ,    false
    )
    {
    //    
    auto_buf<char> dev_driver_id;

    if (!f_GetDeviceProperty(hList, pDevInfoData, SPDRP_DRIVER, REG_SZ, dev_driver_id, strErrorMsg))
        {
        strErrorMsg = "Cannot read driver key: " + strErrorMsg;
        return false;
        }

    //  USB ,    
    HANDLE hHub = f_OpenParentHubByInstance(pDevInfoData->DevInst, strErrorMsg);
    if (INVALID_HANDLE_VALUE == hHub)
        {
        strErrorMsg = "Cannot open parent hub: " + strErrorMsg;
        return false;
        }

    int NumPorts = f_GetHubPortCount(hHub);
    if (NumPorts < 0)
        {
        strErrorMsg = "Cannot get number of ports on hub: " + IntToStr(GetLastError());
        CloseHandle(hHub);
        return false;
        }

    auto_buf<char> ConnInfo;

    for (int nPort = 1; nPort <= NumPorts; nPort++)
        {
        bool connected;
        if (!f_GetHubPortInfo(hHub, nPort, ConnInfo, connected, Info))
            continue;
        //      ,  
        if (!connected)
            continue;
        //       
        auto_buf<char> port_driver_id;
        if (f_GetHubPortDriverKey(hHub, nPort, port_driver_id)
            && !stricmp(port_driver_id.ptr(), dev_driver_id.ptr()))
            { //  
            CloseHandle(hHub);
            strErrorMsg = "";
            // Info  
            return true;
            }
        }

     strErrorMsg = "Cannot find what port the device is attached to";
     CloseHandle(hHub);
     return false;
     }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool USB_GetDeviceInfoByInstance
    (
    DWORD DevInst,                          // ()    (DEVINST)
    AnsiString* pstrDevicePath,             // ()    CreateFile() ( NULL)
    t_USB_ConnInfo* pConnInfo,              // ()    USB ( NULL)
    AnsiString& strErrorMsg                 // ()   ,    false
    )
    {
    SP_DEVICE_INTERFACE_DATA IfaceData;
    SP_DEVINFO_DATA DevInfoData;
    auto_devlist List;

    //   DevInst  DevId
    {
    auto_buf<char> DevId;
    if (!f_GetDevIdByInstance(DevInst, DevId, strErrorMsg))
        return false;

    //   DevId  
    if (!List.open(SetupDiGetClassDevs(&USB_DEVICE_CLSID, DevId.ptr(), NULL, DIGCF_DEVICEINTERFACE)))
        {
        strErrorMsg = "Cannot create device info list";
        return false;
        }
    } // DevId   

    //    (, ..  0)
    ZeroMemory(&IfaceData, sizeof(IfaceData));
    IfaceData.cbSize = sizeof(IfaceData);
    if (!SetupDiEnumDeviceInterfaces(List.Handle(), NULL, &USB_DEVICE_CLSID, 0 /* index */, &IfaceData))
        {
        strErrorMsg = "SetupDiEnumDeviceInterfaces() error " + IntToStr(GetLastError());
        return false;
        }

    ZeroMemory(&DevInfoData, sizeof(DevInfoData));
    DevInfoData.cbSize = sizeof(DevInfoData);
    if (pstrDevicePath)
        {
        auto_buf<SP_DEVICE_INTERFACE_DETAIL_DATA> detail;
        if (!f_GetDeviceByIface(List.Handle(), &IfaceData, &detail, &DevInfoData, strErrorMsg))
            return false;
        *pstrDevicePath = detail.ptr()->DevicePath;
        }
    else
        {
        if (!f_GetDeviceByIface(List.Handle(), &IfaceData, NULL, &DevInfoData, strErrorMsg))
            return false;
        }

    //   / USB,   ;   -  
    if (pConnInfo && !f_GetDeviceConnInfo(List.Handle(), &DevInfoData, *pConnInfo, strErrorMsg))
        {
        pConnInfo->usb_version = 0;
        pConnInfo->speed = USB_Speed_Unknown;
        }

    return true;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool USB_GetDeviceInfoByIface
    (
    HDEVINFO hList,                         // ()    
    PSP_DEVICE_INTERFACE_DATA pIfaceData,   // ()     hList
    AnsiString* pstrDevicePath,             // ()    CreateFile() ( NULL)
    AnsiString* pstrDescription,            // ()  ( NULL)
    DWORD* pDevInst,                        // ()   DEVINST ( NULL)
    AnsiString& strErrorMsg                 // ()   ,    false
    )
    {
    SP_DEVINFO_DATA DevInfoData;

    if ((INVALID_HANDLE_VALUE == hList) || !pIfaceData)
        {
        strErrorMsg = "Invalid parameters";
        return false;
        }

    ZeroMemory(&DevInfoData, sizeof(DevInfoData));
    DevInfoData.cbSize = sizeof(DevInfoData);
    if (pstrDevicePath)
        {
        auto_buf<SP_DEVICE_INTERFACE_DETAIL_DATA> detail;
        if (!f_GetDeviceByIface(hList, pIfaceData, &detail, &DevInfoData, strErrorMsg))
            return false;
        *pstrDevicePath = detail.ptr()->DevicePath;
        }
    else
        {
        if (!f_GetDeviceByIface(hList, pIfaceData, NULL, &DevInfoData, strErrorMsg))
            return false;
        }

    //    SP_DEVINFO_DATA  DEVINST
    if (pDevInst)
        *pDevInst = DevInfoData.DevInst;

    //   ( property)
    if (pstrDescription)
        {
        auto_buf<char> buf;
        if (f_GetDeviceProperty(hList, &DevInfoData, SPDRP_DEVICEDESC, REG_SZ, buf, strErrorMsg))
            {
            *pstrDescription = buf.ptr();
            }
        else
            { //   ,    -  -
            *pstrDescription = "(error_loading_description)";
            }
        }

    return true;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool USB_GetDeviceInfoByPath
    (
    const char* szPath,                     // ()    
    AnsiString* pstrDescription,            // ()  ( NULL)
    DWORD* pDevInst,                        // ()   DEVINST ( NULL)
    AnsiString& strErrorMsg                 // ()   ,    false
    )
    {
    SP_DEVICE_INTERFACE_DATA IfaceData;
    auto_devlist List;

    ZeroMemory(&IfaceData, sizeof(IfaceData));
    IfaceData.cbSize = sizeof(IfaceData);

    //  
    if (!List.open(SetupDiCreateDeviceInfoList(NULL, NULL)))
        {
        strErrorMsg = "Cannot create device info list";
        return false;
        }

    if (!SetupDiOpenDeviceInterface(List.Handle(), szPath, 0, &IfaceData))
        {
        strErrorMsg = "SetupDiOpenDeviceInterface() error " + IntToStr(GetLastError());
        return false;
        }

    return USB_GetDeviceInfoByIface(List.Handle(), &IfaceData, NULL /*    */,
        pstrDescription, pDevInst, strErrorMsg);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool USB_IsKnownDeviceName(const AnsiString& strDevicePath)
    {
    AnsiString s = LowerCase(strDevicePath);
    for (unsigned int i = 0; i < sizeof(KnownUSBDevices) / sizeof(KnownUSBDevices[0]); i++)
        {
        if (s.Pos(KnownUSBDevices[i]))
            return true;
        }
    return false;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
AnsiString USB_ConnInfoString(const t_USB_ConnInfo& info)
    {
    AnsiString res;
    const char* spd_str;
    switch (info.speed)
        {
        case USB_Speed_Low:  spd_str = "low";     break;
        case USB_Speed_Full: spd_str = "full";    break;
        case USB_Speed_High: spd_str = "high";    break;
        default:             spd_str = "unknown"; break;
        }
    if (info.usb_version)
        {
        res = Format("USB %d.%d (%s speed)", ARRAYOFCONST((
            info.usb_version >> 8, (info.usb_version >> 4) & 0x0F, spd_str
            )));
        }
    else
        {
        res = Format("USB (%s speed)", ARRAYOFCONST((spd_str)));
        }
    return res;
    }
/*------------------------------------------------------------------------------------------------*/
/*================================================================================================*/
